# coding=utf-8
# Copyright 2022 The Pax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test lm input generator."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
from paxml import test_helper
from paxml.tasks.lm import input_generator
from praxis import base_hyperparams
from praxis import test_utils
import tensorflow.compat.v2 as tf

instantiate = base_hyperparams.instantiate


class InputTest(test_utils.TestCase):

  # We use first id and seq length as fingerprint to identify each shard
  # has the right elements.
  def _get_first_id_and_lengths(self, batch):
    return batch.labels[:, 1], np.sum(batch.segment_ids, axis=1, dtype=np.int32)

  def test_full(self):
    p = input_generator.TFRecordBertInput.HParams()
    # There are 10 examples in this test data file.
    p.input_file = test_helper.test_src_dir_path('tasks/lm/testdata/tfrecords')
    p.batch_size = 10

    inp = instantiate(p)
    batch = inp.get_next()
    ids, lengths = self._get_first_id_and_lengths(batch)
    expected_ids = np.array(
        [2003, 1996, 1996, 2049, 3748, 1007, 4862, 1996, 2004, 2002],
        dtype=np.int32)
    expected_lengths = np.array([35, 239, 55, 56, 511, 511, 161, 43, 416, 511],
                                dtype=np.int32)
    self.assertArraysEqual(ids, expected_ids)
    self.assertArraysEqual(lengths, expected_lengths)

  def test_remask(self):
    p = input_generator.TFRecordBertInput.HParams()
    # There are 10 examples in this test data file.
    p.input_file = test_helper.test_src_dir_path('tasks/lm/testdata/tfrecords')
    p.batch_size = 10
    p.is_training = True
    p.remask = True
    p.mlm_augmenter.Set(mask_token_id=103, vocab_size=8000)

    inp = instantiate(p)
    batch = inp.get_next()
    ids, lengths = self._get_first_id_and_lengths(batch)
    self.assertEqual(ids.shape, (10,))
    self.assertEqual(lengths.shape, (10,))

  @parameterized.parameters(True, False)
  def test_sharded(self, provide_data_size):
    p = input_generator.TFRecordBertInput.HParams()
    # There are 10 examples in this test data file.
    p.input_file = test_helper.test_src_dir_path('tasks/lm/testdata/tfrecords')
    p.batch_size = 4
    p.eval_data_size = 10 if provide_data_size else 0
    sharded_inputs = [None] * 4
    for i in range(4):
      local_p = p.clone().set(infeed_host_index=i, num_infeed_hosts=4)
      sharded_inputs[i] = instantiate(local_p)

    # This is the same as in test_full() above.
    expected_ids = np.array(
        [2003, 1996, 1996, 2049, 3748, 1007, 4862, 1996, 2004, 2002],
        dtype=np.int32)
    expected_lengths = np.array([35, 239, 55, 56, 511, 511, 161, 43, 416, 511],
                                dtype=np.int32)
    expected_ids = np.reshape(
        np.concatenate(
            [expected_ids, np.array([0] * 6, dtype=np.int32)], axis=0), [4, -1])
    expected_lengths = np.reshape(
        np.concatenate([expected_lengths,
                        np.array([0] * 6, dtype=np.int32)],
                       axis=0), [4, -1])

    for i in [1, 3, 2, 0]:
      # each shard would produce one batch, and then out of range.
      batch = sharded_inputs[i].get_next()
      ids, lengths = self._get_first_id_and_lengths(batch)
      self.assertArraysEqual(ids, expected_ids[i])
      self.assertArraysEqual(lengths, expected_lengths[i])

      with self.assertRaisesRegex(tf.errors.OutOfRangeError, 'End of sequence'):
        sharded_inputs[i].get_next()


if __name__ == '__main__':
  absltest.main()
